// Copyright (C) 2010  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_DATA_Io_HEADER
#define DLIB_DATA_Io_HEADER

#include "data_io/libsvm_io.h"
#include "data_io/image_dataset_metadata.h"
#include "data_io/mnist.h"
#include "data_io/cifar.h"
#include "data_io/arc_agi.h"

#ifndef DLIB_ISO_CPP_ONLY
#include "data_io/load_image_dataset.h"
#endif

#endif // DLIB_DATA_Io_HEADER




